#if !defined(__FLLOPMAT_H)
#define __FLLOPMAT_H
#include "fllopvec.h"

/* subclasses */
#define MATINV          "inv"
#define MATBLOCKDIAG    "blockdiag"
#define MATSUM          "sum"
#define MATPROD         "prod"

typedef enum {MAT_INV_MONOLITHIC, MAT_INV_BLOCKDIAG} MatInvType;
typedef enum {MAT_GATHER_TO_ZERO=0, MAT_GATHER_TO_ALL=1, MAT_GATHER_NONE=-1} MatGatherTarget;
typedef enum {MAT_ORTH_GS, MAT_ORTH_SLEPC, MAT_ORTH_CHOLESKY} MatOrthType;

FLLOP_EXTERN PetscErrorCode MatCreateBlockDiag(MPI_Comm comm, Mat localBlock, Mat *BlockDiag);
FLLOP_EXTERN PetscErrorCode MatTransposeGetMat(Mat N,Mat *A);
FLLOP_EXTERN PetscErrorCode MatCreateSum( MPI_Comm comm,PetscInt nmat,const Mat *mats,Mat *mat);
FLLOP_EXTERN PetscErrorCode MatCreateProd(MPI_Comm comm,PetscInt nmat,const Mat *mats,Mat *mat);
FLLOP_EXTERN PetscErrorCode MatCreateInv(MPI_Comm comm, Mat A, MatInvType invType, Mat *InvA);

PETSC_STATIC_INLINE PetscErrorCode MatCreateNormal_FllopFix(Mat A,Mat *N)
{
  PetscErrorCode ierr;
  ierr = MatCreateNormal(A,N);CHKERRQ(ierr);
  ierr = MatSetUp(*N);CHKERRQ(ierr);
  return 0;
}
#define MatCreateNormal(A,N) MatCreateNormal_FllopFix(A,N)

/* MATINV specific methods */
FLLOP_EXTERN PetscErrorCode MatInvGetMat(Mat InvA, Mat *A);
FLLOP_EXTERN PetscErrorCode MatInvGetKSP(Mat InvA, KSP *ksp);
FLLOP_EXTERN PetscErrorCode MatInvGetPC(Mat InvA, PC *pc);
FLLOP_EXTERN PetscErrorCode MatInvGetType(Mat InvA, MatInvType *type);
FLLOP_EXTERN PetscErrorCode MatInvGetKSPType(Mat InvA, const KSPType *type);
FLLOP_EXTERN PetscErrorCode MatInvGetPCType(Mat InvA, const PCType *type);
FLLOP_EXTERN PetscErrorCode MatInvGetMatSolverPackage(Mat InvA, const MatSolverPackage *pkg);

FLLOP_EXTERN PetscErrorCode MatInvSetMat(Mat InvA, Mat A);
FLLOP_EXTERN PetscErrorCode MatInvSetType(Mat InvA, MatInvType type);
FLLOP_EXTERN PetscErrorCode MatInvSetKSPType(Mat InvA, const KSPType type);
FLLOP_EXTERN PetscErrorCode MatInvSetPCType(Mat InvA, const PCType type);
FLLOP_EXTERN PetscErrorCode MatInvSetMatSolverPackage(Mat InvA, const MatSolverPackage pkg);
FLLOP_EXTERN PetscErrorCode MatInvSetTolerances(Mat Inv, PetscReal rtol, PetscReal abstol, PetscReal dtol,PetscInt maxits);

FLLOP_EXTERN PetscErrorCode MatInvExplicitly(Mat InvA_implicit, PetscBool transpose, Mat InvA_explicit);
FLLOP_EXTERN PetscErrorCode MatInvReset(Mat InvA);
FLLOP_EXTERN PetscErrorCode MatInvSetUp(Mat InvA);

/*   GENERAL Mat   */
FLLOP_EXTERN PetscErrorCode MatFactored(Mat A,PetscBool *flg);
FLLOP_EXTERN PetscErrorCode MatGather(MPI_Comm globcomm, Mat mpiMat, MatGatherTarget target, Mat *seqMat);
FLLOP_EXTERN PetscErrorCode MatCopyLocal(Mat A, Mat B);
FLLOP_EXTERN PetscErrorCode MatPrintInfo(Mat mat);
FLLOP_EXTERN PetscErrorCode MatInvert(Mat mat, MatReuse scall, KSP ksp, Mat *matInv);
FLLOP_EXTERN PetscErrorCode MatIsEye(Mat A, PetscReal tol, PetscInt ntrials, PetscBool *flg);
FLLOP_EXTERN PetscErrorCode MatCreateIdentityOperator(MPI_Comm comm, PetscInt m, PetscInt n, PetscInt N, Mat *I);
FLLOP_EXTERN PetscErrorCode MatCreateTimerWrapper(Mat A, Mat *W);
FLLOP_EXTERN PetscErrorCode MatReplaceByTimerWrapper(Mat *A);
FLLOP_EXTERN PetscErrorCode MatTimerWrapperGetMat(Mat W, Mat *A);
FLLOP_EXTERN PetscErrorCode MatMultEqualTol(Mat A,Mat B,PetscInt n,PetscReal tol,PetscBool  *flg);
FLLOP_EXTERN PetscErrorCode MatMaxEigEst(Mat A, PetscInt iterations, PetscScalar *eigmaxest);
FLLOP_EXTERN PetscErrorCode MatConvertToColVecs(Mat A, Vec *Acols[]);
FLLOP_EXTERN PetscErrorCode MatFilterZeros(Mat A, PetscReal tol, Mat *Af);
FLLOP_EXTERN PetscErrorCode MatConcatHorizontal(PetscInt nmats, Mat mats[], const MatType type, Mat *newmat, IS *map);
FLLOP_EXTERN PetscErrorCode MatInterleaveVertical(MPI_Comm comm,PetscInt nmats, Mat mats[], const MatType type, Mat *newmat, IS *map);
FLLOP_EXTERN PetscErrorCode MatMergeAndDestroy(MPI_Comm comm,Mat *inmat,PetscInt n,Mat *outmat);

/*   ORTHONORMALIZATION   */
FLLOP_EXTERN const char *MatOrthTypes[];
FLLOP_EXTERN PetscErrorCode MatOrthColumns(Mat mat, MatReuse scall, MatOrthType type, Mat *matOrth);

/*   REGULARIZE   */
FLLOP_EXTERN PetscErrorCode MatRegularize(Mat K, Mat R, PetscBool explicitly, Mat *newKreg);

#endif
