#if !defined(__FLLOPPETSCRETRO_H)
#define	__FLLOPPETSCRETRO_H
#include <petscsys.h>
#include <petscksp.h>

#if PETSC_VERSION_MAJOR<3 || PETSC_VERSION_MINOR<1
#error "FLLOP requires PETSc version 3.1 or newer"
#endif

/* petsc-3.2 backward compatibility macros */
#if PETSC_VERSION_MINOR<=2

#undef FLLOP_EXTERN
#if defined(PETSC_USE_EXTERN_CXX) && defined(__cplusplus)
#define FLLOP_EXTERN extern "C"
#else
#define FLLOP_EXTERN extern
#endif

FLLOP_EXTERN PetscErrorCode MatTransposeMatMult(Mat,Mat,MatReuse,PetscReal,Mat*);

PETSC_STATIC_INLINE PetscErrorCode VecCreateSeqWithArray_3_2(MPI_Comm comm,PetscInt bs,PetscInt n,const PetscScalar array[],Vec *V)
{
  PetscErrorCode ierr;
  ierr = VecCreateSeqWithArray(comm,n,array,V);CHKERRQ(ierr);
  ierr = VecSetBlockSize(*V,bs);CHKERRQ(ierr);
  return 0;
}
#define VecCreateSeqWithArray(comm,bs,n,array,V) VecCreateSeqWithArray_3_2(comm,bs,n,array,V)

PETSC_STATIC_INLINE PetscErrorCode VecCreateMPIWithArray_3_2(MPI_Comm comm,PetscInt bs,PetscInt n,PetscInt N,const PetscScalar array[],Vec *V)
{
  PetscErrorCode ierr;
  ierr = VecCreateMPIWithArray(comm,n,N,array,V);CHKERRQ(ierr);
  ierr = VecSetBlockSize(*V,bs);CHKERRQ(ierr);
  return 0;
}
#define VecCreateMPIWithArray(comm,bs,n,N,array,V) VecCreateMPIWithArray_3_2(comm,bs,n,N,array,V)

#define PetscObjectTypeCompare(obj,type_name,same)                     PetscTypeCompare(obj,type_name,same)
#define MatCreateDense(comm,m,n,M,N,data,A)                            MatCreateMPIDense(comm,m,n,M,N,data,A)
#define MatCreateMPIAIJConcatenateSeqAIJ(comm,inmat,n,scall,outmat)    MatMerge(comm,inmat,n,scall,outmat)
#endif

/* petsc-3.1 backward compatibility macros */
#if PETSC_VERSION_MINOR==1

#define MATSOLVERPETSC MAT_SOLVER_PETSC
#define PetscClassId PetscCookie
#define PetscClassIdRegister(name,oclass) PetscCookieRegister(name,oclass)
#define MAT_CLASSID MAT_COOKIE
#define VEC_CLASSID VEC_COOKIE
#define KSP_CLASSID KSP_COOKIE
#define PC_CLASSID  PC_COOKIE
#define MATORDERINGNATURAL MATORDERING_NATURAL

#undef FLLOP_SETERRQ
#undef FLLOP_SETERRQ1
#undef FLLOP_SETERRQ2
#undef FLLOP_SETERRQ3
#undef FLLOP_SETERRQ4
#undef FLLOP_SETERRQ5
#define FLLOP_SETERRQ( comm,n,s) SETERRQ(n,s)
#define FLLOP_SETERRQ1(comm,n,s,a1) SETERRQ1(n,s,a1)
#define FLLOP_SETERRQ2(comm,n,s,a1,a2) SETERRQ2(n,s,a1,a2)
#define FLLOP_SETERRQ3(comm,n,s,a1,a2,a3) SETERRQ3(n,s,a1,a2,a3)
#define FLLOP_SETERRQ4(comm,n,s,a1,a2,a3,a4) SETERRQ4(n,s,a1,a2,a3,a4)
#define FLLOP_SETERRQ5(comm,n,s,a1,a2,a3,a4,a5) SETERRQ5(n,s,a1,a2,a3,a4,a5)

#undef PetscHeaderCreate
#define PetscHeaderCreate(h,tp,pops,cook,t,class_name,descr,mansec,com,des,vie)	\
  (PetscNew(struct tp,&(h)) ||						\
   PetscNew(PetscOps,&(((PetscObject)(h))->bops)) ||			\
   PetscNew(pops,&((h)->ops)) ||					\
   PetscHeaderCreate_Private((PetscObject)h,cook,t,class_name,com,(PetscObjectFunction)des,(PetscObjectViewerFunction)vie) || \
   PetscLogObjectCreate(h) ||						\
   PetscLogObjectMemory(h, sizeof(struct tp) + sizeof(PetscOps) + sizeof(pops)))

#undef PetscHeaderDestroy
#define PetscHeaderDestroy(h)			   \
  (PetscLogObjectDestroy((PetscObject)(*h)) ||	   \
   PetscComposedQuantitiesDestroy((PetscObject)*h) || \
   PetscHeaderDestroy_Private((PetscObject)(*h)) || \
   PetscFree((*h)->ops) ||			   \
   PetscFree(*h))


#if defined(PETSC_USE_REAL___FLOAT128)
FLLOP_EXTERN  MPI_Op MPIU_MAX;
FLLOP_EXTERN  MPI_Op MPIU_MIN;
#else
#define MPIU_MAX MPI_MAX
#define MPIU_MIN MPI_MIN
#endif

#define PetscValidLogicalCollectiveScalar(a,b,c)                        \
  do {                                                                  \
    PetscErrorCode _7_ierr;                                             \
    PetscReal b1[2] = {-PetscRealPart(b),PetscRealPart(b)},b2[2];       \
    _7_ierr = MPI_Allreduce(b1,b2,2,MPIU_REAL,MPIU_MAX,((PetscObject)a)->comm);CHKERRQ(_7_ierr); \
    if (-b2[0] != b2[1]) SETERRQ1(PETSC_ERR_ARG_WRONG,"Scalar value must be same on all processes, argument # %d",c); \
  } while (0)

#define PetscValidLogicalCollectiveReal(a,b,c)                          \
  do {                                                                  \
    PetscErrorCode _7_ierr;                                             \
    PetscReal b1[2] = {-b,b},b2[2];                                     \
    _7_ierr = MPI_Allreduce(b1,b2,2,MPIU_REAL,MPIU_MAX,((PetscObject)a)->comm);CHKERRQ(_7_ierr); \
    if (-b2[0] != b2[1]) SETERRQ1(PETSC_ERR_ARG_WRONG,"Real value must be same on all processes, argument # %d",c); \
  } while (0)

#define PetscValidLogicalCollectiveInt(a,b,c)                           \
  do {                                                                  \
    PetscErrorCode _7_ierr;                                             \
    PetscInt b1[2] = {-b,b},b2[2];                                      \
    _7_ierr = MPI_Allreduce(b1,b2,2,MPIU_INT,MPI_MAX,((PetscObject)a)->comm);CHKERRQ(_7_ierr); \
    if (-b2[0] != b2[1]) SETERRQ1(PETSC_ERR_ARG_WRONG,"Int value must be same on all processes, argument # %d",c); \
  } while (0)

#define PetscValidLogicalCollectiveBool(a,b,c)                         \
  do {                                                                  \
    PetscErrorCode _7_ierr;                                             \
    PetscMPIInt b1[2] = {-(PetscMPIInt)b,(PetscMPIInt)b},b2[2];         \
    _7_ierr = MPI_Allreduce(b1,b2,2,MPI_INT,MPI_MAX,((PetscObject)a)->comm);CHKERRQ(_7_ierr); \
    if (-b2[0] != b2[1]) SETERRQ1(PETSC_ERR_ARG_WRONG,"Bool value must be same on all processes, argument # %d",c); \
  } while (0)

#define PetscValidLogicalCollectiveEnum(a,b,c)                          \
  do {                                                                  \
    PetscErrorCode _7_ierr;                                             \
    PetscMPIInt b1[2] = {-(PetscMPIInt)b,(PetscMPIInt)b},b2[2];         \
    _7_ierr = MPI_Allreduce(b1,b2,2,MPI_INT,MPI_MAX,((PetscObject)a)->comm);CHKERRQ(_7_ierr); \
    if (-b2[0] != b2[1]) SETERRQ1(PETSC_ERR_ARG_WRONG,"Enum value must be same on all processes, argument # %d",c); \
  } while (0)

#define PetscBool PetscTruth

#define VecDestroy(obj)         (*obj) ? VecDestroy(*obj) : 0; *obj=0;
#define MatDestroy(obj)         (*obj) ? MatDestroy(*obj) : 0; *obj=0;
#define KSPDestroy(obj)         (*obj) ? KSPDestroy(*obj) : 0; *obj=0;
#define PCDestroy(obj)          (*obj) ? PCDestroy(*obj)  : 0; *obj=0;
#define VecScatterDestroy(obj)  (*obj) ? VecScatterDestroy(*obj) : 0; *obj=0;
#define PetscViewerDestroy(obj) (*obj) ? PetscViewerDestroy(*obj) : 0; *obj=0;
#define ISDestroy(obj)          (*obj) ? ISDestroy(*obj) : 0; *obj=0;
#define PetscRandomDestroy(obj) (*obj) ? PetscRandomDestroy(*obj) : 0; *obj=0;
#define VecDestroyVecs(m,obj)   VecDestroyVecs(*obj,m);
#define PetscFListFind(fl,comm,name,searchlibraries,r)  PetscFListFind(fl,comm,name,r)
#define MatMPIAIJGetLocalMat(mat,reuse,matloc) MatGetLocalMat(mat,reuse,matloc)

PETSC_STATIC_INLINE PetscErrorCode MatShellGetContext_3_1(Mat mat,void *ctx) {
  PetscErrorCode ierr;
  ierr = MatShellGetContext(mat, &ctx);CHKERRQ(ierr);
  return 0;
}
#define MatShellGetContext(mat, ctx) MatShellGetContext_3_1(mat, ctx)

typedef enum { PETSC_COPY_VALUES, PETSC_OWN_POINTER, PETSC_USE_POINTER} PetscCopyMode;
PETSC_STATIC_INLINE PetscErrorCode ISCreateGeneral_3_1(MPI_Comm comm,PetscInt n,PetscInt idx[],PetscCopyMode mode,IS *is) {
  PetscErrorCode ierr;
  switch(mode) {
    case PETSC_COPY_VALUES: ierr = ISCreateGeneral(comm,n,idx,is);CHKERRQ(ierr); break;
    case PETSC_OWN_POINTER: ierr = ISCreateGeneralNC(comm,n,idx,is);CHKERRQ(ierr); break;
    default:                ierr = ISCreateGeneralWithArray(comm,n,idx,is);CHKERRQ(ierr); break;
  }
  return 0;
}
#define ISCreateGeneral(comm,n,idx,mode,is) ISCreateGeneral_3_1(comm,n,idx,mode,is)

FLLOP_EXTERN PetscErrorCode (*PetscOptionsGetBool)(const char[],const char [],PetscBool*,PetscBool*);
FLLOP_EXTERN PetscErrorCode (*PetscOptionsBool)(const char[],const char[],const char[],PetscBool,PetscBool*,PetscBool*);
FLLOP_EXTERN PetscErrorCode (*MatMPIAIJGetLocalMat)(Mat,MatReuse,Mat*);

#endif

#endif
