#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "layplate.h"
#include "global.h"
#include "vecttens.h"
#include "intpoints.h"
#include "matrix.h"
#include "vector.h"
#include "tensor.h"
#include "elastisomat.h"
#include "strretalg.h"
#include "alias.h"
#include "element.h"

/**
  This constructor inializes attributes to zero values.
*/
layplate::layplate (void)
{
  
}


/**
  The destructor
*/
layplate::~layplate (void)
{

}


/**
  This function reads material parameters from the opened text file given
  by the parameter in.

  @param in - pointer to the opened text file

  10/2014   J. Fiedler
*/
void layplate::read (XFILE *in)
{
  long i,j;
//  long eid = Mm->elip[ipp];
  
  xfscanf (in,"%k%ld", "num_lay", &nl);
/*  if (Mc->give_num_lay(eid)!=nl){
    print_err("wrong number of layers defined", __FILE__, __LINE__, __func__);
  }
*/    
  tm   = new mattype*[nl];
  idm  = new long*[nl];
  nm   = new long[nl];
  
  for (i=0;i<nl;i++){                       // cyklus pres vrstvy
    xfscanf (in,"%ld", &nm[i]);             // nacteni poctu materialu na vrstve
    tm[i]  = new mattype[nm[i]];            // alokovani tm
    idm[i] = new long[nm[i]];               // alokovani idm

    for (j=0;j<nm[i];j++){
      xfscanf(in, "%k%m%k%ld", "mattype", &mattype_kwdset, (int*)(tm[i]+j), "num_inst", idm[i]+j);
      idm[i][j]--;
    }
  }
  xfscanf (in,"%k%ld", "num_iter", &nli);
  xfscanf (in,"%k%lf", "error", &err);

  allocm(3,3,c);
  allocm(3,3,cn);
}



/**
  This function computes material stiffness %matrix considering layer stiffnesses.

  @param d   - material stiffness matrix
  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  11/2014   J. Fiedler
*/
void layplate::matstiff (matrix &d, long ipp,long ido)
{
  long i,j,jj;
  double *th;
  double *zet;
  matrix e(3,3);
  matrix dn(3,3);
  matrix dnm(3,3);
  matrix dm(3,3);
  matrix pomat(3,3);
  matrix pommat(3,3);
  long eid = Mm->elip[ipp];
  double *k;

  th   = Mc->give_layer_thicke(eid);
  zet  = Mc->give_layer_zcoord(eid);
  
  fillm(0,e);
  fillm(0,d);
  backup(ipp,k,0);
  for (i=0;i<nl;i++){
    Mm->ip[ipp].tm  = tm [i];     // mattype aktualni vrstvy
    Mm->ip[ipp].idm = idm[i];     // index sady mat. vl. aktualni sady
    Mm->ip[ipp].nm  = nm [i];
  
    Mm->elmatstiff (e,ipp);
    
    for (j=0;j<3;j++){
      for (jj=0;jj<3;jj++){
        dn[j][jj] += th[i]*e[j][jj];
      }
    }
    
    for (j=0;j<3;j++){
      for (jj=0;jj<3;jj++){
        dnm[j][jj] += zet[i]*th[i]*e[j][jj];
      }
    }
    
    for (j=0;j<3;j++){
      for (jj=0;jj<3;jj++){
        dm[j][jj] += th[i]*zet[i]*zet[i]*e[j][jj];
      }
    }
  }
  restore_values(ipp,k,0);
// vypocet matice tuhosti za predpokladu N = 0
  invm(dn,c,1.0e-20);
  cmulm(-1.0,dnm,pomat);
  mxm(c,pomat,cn);  // -> priprava pro vypocet eps v ramci nlstresses
  mxm(pomat,c,pommat);
  mxm(pommat,dnm,pomat);
  addm(pomat,dm,d);

}


/**
  This function computes stresses at given integration point ipp,
  depending on the reached strains. The stress and the other attribute of
  given integration point is actualized.

  @param ipp - integration point number in the mechmat ip array.
  @param im  - index of material type for given ip
  @param ido - index of internal variables for given material in the ipp other array

  @return The function updates the stress array of the given integration point.

  10/2014 J. Fiedler
*/
void layplate::nlstresses (long ipp, long im, long ido)
{
  long i,j,nc = Mm->ip[ipp].ncompstr;
  long eid = Mm->elip[ipp];
  double *k;          // vektor krivosti
  double *layth;      // ukazatel na pole tloustek
  double *layz;       // ukazatel na pole z-tovych souradnic
  double th;
  vector intfor(nc);  // vnitrni sily - momenty
  vector df(3);       // vektor nevyrovnanych sil
  vector eps(3);       // epsilony
  vector dpeps(3);
  vector deps(3);
  double *asd;
  double normf;
  FILE *out;
  char fname[14] = "kontrola.dat";

  if (Mp->istep==97 && Mp->jstep==2 && Mm->elip[ipp]==3 && ipp==10)
    out = fopen(fname, "wt");
  //  ulozeni potrebnych ukazatelu
  layth = Mc->give_layer_thicke(eid);
  layz  = Mc->give_layer_zcoord(eid);
  for (i=0;i<3;i++)
    dpeps[i] = Mm->ip[ipp].eqother[i];   // naposled dosazeny posun neutralni osy

  backup (ipp,k,1);                   // zaloha informaci na integracnim bode

  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      eps[i] += cn[i][j]*k[j];

  for (i=0;i<3;i++)
    eps[i] += dpeps[i];
  
  for (i=0;i<nli;i++)                 // iteracni cyklus pro rovnovahu norm. sily na prurezu
  {
    stress_calc(df,eps,k,intfor,layz,layth,ipp,ido,out);
    if (Mp->istep==97 && Mp->jstep==2 && Mm->elip[ipp]==3 && ipp==10)
      fprintf(out,"  df= %lf\n",df[0]);
    normf = normv(df);
    if (normf<err)
      break;
    mxv(c,df,deps);
    for (j=0;j<3;j++)
      eps[j] -= deps[j];
    for (j=0;j<3;j++)
      Mm->ip[ipp].other[j] -= deps[j];    
    if (i==nli-1 && normf>err)
      {
        print_err("cross-section equilibrium has not been reached (istep=%ld,jstep=%ld,nordf=%le, el=%ld, ipp=%ld)",
                  __FILE__, __LINE__, __func__, Mp->istep, Mp->jstep, normf, Mm->elip[ipp], ipp);
        fprintf (Out,"\n cross-section equilibrium has not been reached");
      }
  }
  asd = Mm->ip[ipp].other;
  if (Mp->istep==97 && Mp->jstep==2 && Mm->elip[ipp]==3 && ipp==10)
    fclose(out);
    
  Mm->storestress(0, ipp, intfor);
  restore_values (ipp,k,1);         // vraceni na puvodni hodnoty
}


/**
  This function evaluates the size of eqother array.

  @param ipp - integration point number in the mechmat ip array.

  @return The function returns the size of eqother array depending of number of layers
          and their material type.

  10/2014 J. Fiedler
*/
long layplate::compeqother(long ipp)  // -> zavolat tuhle funkci v mechmatu ve funkci givencomeqother
{
  long sizeeqother=0;
  long i;
  mattype *btm = Mm->ip[ipp].tm;
  Mm->ip[ipp].ncompstr = 4;

  // cyklus pres vrstvy
  for (i=0;i<nl;i++){
    Mm->ip[ipp].tm  = tm [i];
    sizeeqother += Mm->givencompeqother(ipp,0)+3;
  }

  Mm->ip[ipp].tm  = btm;  // vraci mattype na puvodni layplate
  Mm->ip[ipp].ncompstr = 3;
  
  return sizeeqother+3;
}


/**
  This function evaluates the size of other array.

  @param ipp - integration point number in the mechmat ip array.

  @return The function returns the size of other array depending of number of layers
          and their material type.

  10/2014 J. Fiedler
*/
long layplate::compother(long ipp)  // -> zavolat tuhle funkci v mechmatu ve funkci givencompother
{
  long sizeother=0;
  long i;
  mattype *btm = Mm->ip[ipp].tm;
  Mm->ip[ipp].ncompstr = 4;

  // cyklus pres vrstvy
  for (i=0;i<nl;i++){
    Mm->ip[ipp].tm  = tm[i];
    sizeother += Mm->givencompother(ipp,0)+3;
  }

  Mm->ip[ipp].tm  = btm;  // vraci mattype na puvodni layplate
  Mm->ip[ipp].ncompstr = 3;
  
  return sizeother+3;
}


/**
  Function backs up some informations stored on integration point and prepares arrays for computation.

  @param ipp - integration point number in the mechmat ip array.

  10/2014 J. Fiedler
*/
void layplate::backup(long ipp, double *&k, long j)
{
  bcup.tm   = Mm->ip[ipp].tm;
  bcup.idm  = Mm->ip[ipp].idm;
  bcup.nm   = Mm->ip[ipp].nm;
  bcup.ssst = Mm->ip[ipp].ssst;
  bcup.ncompstr = Mm->ip[ipp].ncompstr;
  Mm->ip[ipp].ssst = planestress;
  Mm->ip[ipp].ncompstr = 4;

  if (j==1)
  {
    bcup.stress = Mm->ip[ipp].stress;
    k = Mm->ip[ipp].strain;
    Mm->ip[ipp].alloc_strains(1);   // znovualokovani podle noveho stress-strain state
    Mm->ip[ipp].alloc_stresses(1);
    Mm->ip[ipp].clean_strains(1);
    Mm->ip[ipp].clean_stresses(1);
  }
}

/**
  Function restores original informations stored on integration point.

  @param ipp - integration point number in the mechmat ip array.

  10/2014 J. Fiedler
*/
void layplate::restore_values(long ipp, double *k, long j)
{
  long i;
  
  Mm->ip[ipp].tm       = bcup.tm;
  Mm->ip[ipp].idm      = bcup.idm;
  Mm->ip[ipp].nm       = bcup.nm;
  Mm->ip[ipp].ssst     = bcup.ssst;
  Mm->ip[ipp].ncompstr = bcup.ncompstr;

  if (j==1)
  {
    delete [] Mm->ip[ipp].strain;
    Mm->ip[ipp].strain = k;
    for (i=0; i<3; i++)
      bcup.stress[i] = Mm->ip[ipp].stress[i];
    delete [] Mm->ip[ipp].stress;
    Mm->ip[ipp].stress = bcup.stress;
  }
}

/**
  Function calculates stresses on all layers and resulting internal forces
*/
void layplate::stress_calc(vector &df, vector &eps, double *k, vector &intfor, double *layz, double *layth, long ipp, long ido, FILE *&out)
{
  long i,j,nc = Mm->ip[ipp].ncompstr;
  long neq;
  
  for (i=0;i<3;i++){
    df[i] = 0.0;
  }
  
  ido = 3;

  for (i=0;i<nl;i++)
  {
    Mm->ip[ipp].tm  = tm [i];     // mattype aktualni vrstvy
    Mm->ip[ipp].idm = idm[i];     // index sady mat. vl. aktualni sady
    Mm->ip[ipp].nm  = nm [i];

    for (j=0;j<3;j++){           // cyklus pres slozky deformaci -> vypocet deformace pro prislusnou vrstvu
      Mm->ip[ipp].strain[j] = k[j]*layz[i]+eps[j];
    }
    
    Mm->computenlstresses(ipp,0,ido);
    
    for (j=0;j<3;j++){
      intfor[j] += Mm->ip[ipp].stress[j]*layth[i]*layz[i];  // nascitavani momentu
    }
        
    neq = Mm->givencompeqother(ipp,0);
    for (j=0;j<3;j++){
      Mm->ip[ipp].other[ido+neq+j] = Mm->ip[ipp].stress[j];       // ukladani napeti na vrstve do pole other
    }

    if (Mp->istep==97 && Mp->jstep==2 && Mm->elip[ipp]==3 && ipp==10)
      fprintf(out," %lf %lf",Mm->ip[ipp].stress[0],Mm->ip[ipp].strain[0]);

    for (j=0;j<3;j++)
      df[j] += Mm->ip[ipp].stress[j]*layth[i];

    ido += neq+3;      // prepocet ido pro nasledujici vrstvu
  }
}


/**
  This function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.
*/
void layplate::updateval (long ipp, long im, long ido)
{
  long i,n = Mm->givencompeqother(ipp,im);

  for (i=0;i<n;i++)
    Mm->ip[ipp].eqother[ido+i]=Mm->ip[ipp].other[ido+i];
}

